/**
 * This module provides interface routines to the AVR uart
 * \author 2001, R O SoftWare, modyfied by Vladimir Ermakov
 * \date Sun Nov 25 19:01:55 MSK 2007
 */

/**
 */
#include "usart.h"

#ifdef USART0_INT_MODE
#ifndef USART0_TX_INT_MODE
#define USART0_TX_INT_MODE
#endif
#ifndef USART0_RX_INT_MODE
#define USART0_RX_INT_MODE
#endif
#endif

#ifdef USART1_INT_MODE
#ifndef USART1_TX_INT_MODE
#define USART1_TX_INT_MODE
#endif
#ifndef USART1_RX_INT_MODE
#define USART1_RX_INT_MODE
#endif
#endif

/**
 * various processors have different register names
 * we'll use the most encompassing
 * @{
 */
#ifdef UCSRA
#define UCSR0A  UCSRA
#define UCSR0B  UCSRB
#elif USR
#define UCSR0A  USR
#define UCSR0B  UCR
#endif
#ifndef UDR0
#define UDR0    UDR
#define UBRR0   UBRR

#define RXC0    RXC
#define TXC0    TXC
#define UDRE0   UDRE
#define FE0     FE
#define OVR0    OVR
#define U2X0    U2X

#define RXCIE0  RXCIE
#define TXCIE0  TXCIE
#define UDRIE0  UDRIE
#define RXEN0   RXEN
#define TXEN0   TXEN
#endif
#ifndef OVR0
#define OVR0    DOR0
#define OVR1    DOR1
#endif
#ifndef UBRR0H
#define UBRR0H	UBRRH
#define UBRR0L	UBRRL
#define USART0_RX_vect  	USART_RXC_vect
#define USART0_UDRE_vect	USART_UDRE_vect
#endif
#ifndef OVR
#define OVR DOR
#endif
///@}

#if USART0_SUPPORT
namespace usart0
{
	#ifdef USART0_RX_INT_MODE
	static uint8_t  rx_buffer[USART0_RX_BUFFER_SIZE];
	static uint16_t rx_insert_idx, rx_extract_idx;
	#endif

	#ifdef USART0_TX_INT_MODE
	static uint8_t  tx_buffer[USART0_TX_BUFFER_SIZE];
	static uint16_t tx_insert_idx, tx_extract_idx;
	#endif

	// This function initializes the USART
	void init(uint16_t baud)
	{
		// disable the USART
		UCSR0B = 0x00;
		UCSR0A = 0x00;

		if(baud & DOUBLE_SPEED_BIT)
			UCSR0A |= (1<<U2X0);

		// load the baudrate divisor register
		UBRR0L = (uint8_t)baud;

		// output the upper four bits of the baudrate divisor
		UBRR0H = (UBRR0H & 0xF0) | ((baud >> 8) & 0x0F);

		// enable the USART0 transmitter & receiver
		UCSR0B = (1<<TXEN0) | (1<<RXEN0);

		#ifdef USART0_TX_INT_MODE
		tx_extract_idx = tx_insert_idx = 0;
		#endif

		#ifdef USART0_RX_INT_MODE
		rx_extract_idx = rx_insert_idx = 0;

		// enable receiver interrupts
		UCSR0B |= (1<<RXCIE0);
		#endif
	}

	// This function puts a character into the USART output queue for
	// transmission.
	int16_t putch(char ch)
	{
		#ifdef USART0_TX_INT_MODE
		uint16_t temp;

		temp = (tx_insert_idx + 1) % USART0_TX_BUFFER_SIZE;

		if (temp == tx_extract_idx)
			return -1; // no room

		// add to queue
		tx_buffer[tx_insert_idx] = (uint8_t)ch;
		tx_insert_idx = temp;
		UCSR0B |= (1<<UDRIE0); // enable TX interrupts

		#else

		while(bit_is_clear(UCSR0A, UDRE0)) // wait for TX buffer to empty
			wdt_reset();

		UDR0 = ch;

		#endif // USART0_TX_INT_MODE

		return (uint8_t)ch;
	}

	// This function gets the available space in the transmit queue
	uint16_t space(void)
	{
		#ifdef USART0_TX_INT_MODE
		int space;

		if((space = (tx_extract_idx - tx_insert_idx)) <= 0)
			space += USART0_TX_BUFFER_SIZE;

		return (uint16_t)(space - 1);
		#else
		return 0xffff;
		#endif // USART0_TX_INT_MODE
	}

	// This function writes a NULL terminated 'string' to the USART output
	// queue, returning a pointer to the next character to be written.
	const char *puts(const char *string)
	{
	register char ch;

	while((ch = *string) && (putch(ch) >= 0))
		string++;

	return string;
	}

	// This function writes 'count' characters from 'buffer' to the USART0
	// output queue.
	int write(const char *buffer, uint16_t count)
	{
		#ifdef USART0_TX_INT_MODE
		if(count > space())
			return -1;
		#endif // USART0_TX_INT_MODE
		while(count && (putch(*buffer++) >= 0))
			count--;

		return (count ? -2 : 0);
	}

	// This function returns the status of the USART transmit data
	// registers.
	int tx_empty(void)
	{
		return bit_is_set(UCSR0A, TXC0);
	}

	// This function removes all characters from the USART transmit queue
	// (without transmitting them).
	void tx_flush(void)
	{
		#ifdef USART0_TX_INT_MODE
		/* "Empty" the transmit buffer. */
		UCSR0B &= ~(1<<UDRIE0); // disable TX interrupts
		tx_insert_idx = tx_extract_idx = 0;
		UCSR0B |= (1<<UDRIE0); // enable TX interrupts
		#endif // USART0_TX_INT_MODE
	}

	// This function gets a character from the USART receive queue
	int16_t getch(void)
	{
		uint8_t ch;

		#ifdef USART0_RX_INT_MODE
		if(rx_insert_idx == rx_extract_idx) // check if character is available
			return -1;

		ch = rx_buffer[rx_extract_idx++]; // get character, bump pointer
		rx_extract_idx %= USART0_RX_BUFFER_SIZE; // limit the pointer
		return ch;
		#else
		if(bit_is_set(UCSR0A, RXC0)) // check if character is available
		{
			ch = UDR0; // get the character

			if( !(UCSR0A & ((1<<FE0) | (1<<OVR0))) ) // check for errors
				return ch; // none, return character
		}

		return -1; // no valid cahracter
		#endif // USART0_RX_INT_MODE
	}
} // namespace usart0

#ifdef USART0_RX_INT_MODE
// uart receive isr
ISR(USART0_RX_vect)
{
	uint16_t temp;

	// check status register for receive errors
	if(UCSR0A & ((1<<FE0) | (1<<OVR0)))
	{
		// this character is no good
		temp = UDR0; // dummy read
		return;
	}

	temp = (usart0::rx_insert_idx + 1) % USART0_RX_BUFFER_SIZE;
	usart0::rx_buffer[usart0::rx_insert_idx] = UDR0; // read the character

	if(temp != usart0::rx_extract_idx)
		usart0::rx_insert_idx = temp;
}
#endif // USART0_RX_INT_MODE

#ifdef USART0_TX_INT_MODE
// uart transmit isr
ISR(USART0_UDRE_vect)
{
	if (usart0::tx_insert_idx != usart0::tx_extract_idx)
	{
		UDR0 = usart0::tx_buffer[usart0::tx_extract_idx++];
		usart0::tx_extract_idx %= USART0_TX_BUFFER_SIZE;
	}
	else
		UCSR0B &= ~(1<<UDRIE0); // disable TX interrupts
}
#endif // USART0_TX_INT_MODE
#endif // USART0_SUPPORT


#if USART1_SUPPORT
namespace usart1
{
	#ifdef USART1_RX_INT_MODE
	static uint8_t  rx_buffer[USART1_RX_BUFFER_SIZE];
	static uint16_t rx_insert_idx, rx_extract_idx;
	#endif

	#ifdef USART1_TX_INT_MODE
	static uint8_t  tx_buffer[USART1_TX_BUFFER_SIZE];
	static uint16_t tx_insert_idx, tx_extract_idx;
	#endif

	// This function initializes the USART
	void init(uint16_t baud)
	{
		// disable the USART
		UCSR1B = 0x00;
		UCSR1A = 0x00;

		// check if double speed operation is requested
		if(baud & DOUBLE_SPEED_BIT)
			UCSR1A |= (1<<U2X1);

		// load the baudrate divisor register
		UBRR1L = (uint8_t)baud;

		// output the upper four bits of the baudrate divisor
		UBRR1H = (UBRR1H & 0x0F) | ((uint8_t)(baud >> 4) & 0xF0);

		// enable the USART1 transmitter & receiver
		UCSR1B = (1<<TXEN1) | (1<<RXEN1);

		#ifdef USART1_TX_INT_MODE
		tx_extract_idx = tx_insert_idx = 0;
		#endif // USART1_TX_INT_MODE

		#ifdef USART1_RX_INT_MODE
		rx_extract_idx = rx_insert_idx = 0;

		// enable receiver interrupts
		UCSR1B |= (1<<RXCIE1);
		#endif // USART1_RX_INT_MODE
	}

	// This function puts a character into the USART output queue for
	// transmission.
	int16_t putch(char ch)
	{
		#ifdef USART1_TX_INT_MODE
		uint16_t temp;

		temp = (tx_insert_idx + 1) % USART1_TX_BUFFER_SIZE;

		if(temp == tx_extract_idx)
			return -1; // no room

		// add to queue
		tx_buffer[tx_insert_idx] = (uint8_t)ch;
		tx_insert_idx = temp;
		UCSR1B |= (1<<UDRIE1); // enable TX interrupts

		#else

		while(bit_is_clear(UCSR1A, UDRE1)) // wait for TX buffer to empty
			wdt_reset();

		UDR1 = ch;

		#endif // USART1_TX_INT_MODE

		return (uint8_t)ch;
	}

	// This function gets the available space in the transmit queue
	uint16_t space(void)
	{
		#ifdef USART1_TX_INT_MODE
		int space;

		if((space = (tx_extract_idx - tx_insert_idx)) <= 0)
			space += USART1_TX_BUFFER_SIZE;

		return (uint16_t)(space - 1);
		#else
		return 0xffff;
		#endif // USART1_TX_INT_MODE
	}

	// This function writes a NULL terminated 'string' to the USART output
	// queue, returning a pointer to the next character to be written.
	const char *puts(const char *string)
	{
		register char ch;

		while((ch = *string) && (putch(ch) >= 0))
			string++;

		return string;
	}

	// This function writes 'count' characters from 'buffer' to the USART1
	// output queue.
	int write(const char *buffer, uint16_t count)
	{
		#ifdef USART1_TX_INT_MODE
		if(count > space())
			return -1;
		#endif // USART1_TX_INT_MODE
		while(count && (putch(*buffer++) >= 0))
			count--;

		return (count ? -2 : 0);
	}

	// This function returns the status of the USART transmit data
	// registers.
	int tx_empty(void)
	{
		return bit_is_set(UCSR1A, TXC1);
	}

	// This function removes all characters from the USART transmit queue
	// (without transmitting them).
	void tx_flush(void)
	{
		#ifdef USART1_TX_INT_MODE
		/* "Empty" the transmit buffer. */
		UCSR1B &= ~(1<<UDRIE1); // disable TX interrupts
		tx_insert_idx = tx_extract_idx = 0;
		UCSR1B |= (1<<UDRIE1); // enable TX interrupts
		#endif // USART1_TX_INT_MODE
	}

	// This function gets a character from the USART receive queue
	int16_t getch(void)
	{
		uint8_t ch;

		#ifdef USART1_RX_INT_MODE
		if(rx_insert_idx == rx_extract_idx) // check if character is available
			return -1;

		ch = rx_buffer[rx_extract_idx++]; // get character, bump pointer
		rx_extract_idx %= USART1_RX_BUFFER_SIZE; // limit the pointer
		return ch;
		#else
		if (bit_is_set(UCSR1A, RXC1))// check if character is available
		{
			ch = UDR1; // get the character

			if( !(UCSR1A & ((1<<FE1) | (1<<OVR1))) ) // check for errors
				return ch; // none, return character
		}

		return -1; // no valid cahracter
		#endif // USART1_RX_INT_MODE
	}
} // namespace usart1

#ifdef USART1_RX_INT_MODE
// uart receive isr
ISR(USART1_RX_vect)
{
	uint16_t temp;

	// check status register for receive errors
	if(UCSR1A & ((1<<FE1) | (1<<OVR1)))
		{
		// this character is no good
		temp = UDR1; // dummy read
		return;
		}

	temp = (rx_insert_idx + 1) % USART1_RX_BUFFER_SIZE;
	rx_buffer[rx_insert_idx] = UDR1; // read the character

	if (temp != rx_extract_idx)
		rx_insert_idx = temp;
}
#endif // USART1_RX_INT_MODE

#ifdef USART1_TX_INT_MODE
// uart transmit isr
ISR(USART1_UDRE_vect)
{
	if(tx_insert_idx != tx_extract_idx)
	{
		UDR1 = tx_buffer[tx_extract_idx++];
		tx_extract_idx %= USART1_TX_BUFFER_SIZE;
	}
	else
		UCSR1B &= ~(1<<UDRIE1); // disable TX interrupts
}
#endif // USART1_TX_INT_MODE
#endif // USART1_SUPPORT
