/**
 * This module provides interface routines to the AVR usart
 * \author 2001, R O SoftWare, modyfied by Vladimir Ermakov
 * \date Sun Nov 25 19:01:55 MSK 2007
 * \defgroup usart_drv <usart.h>: AVR USART API
 * \addtogroup usart_drv
 * \code #include "usart/usart.h" \endcode
 * @{
 */

/**
 */
#ifndef USART_H
#define USART_H

#include <inttypes.h>
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

//-----------------------------------------------------------------------------
/**
 * config
 * \defgroup usart_drv_cfg <usart.h>: USART Config
 * \addtogroup usart_drv_cfg
 * @{
 */

/// non-zero to enable usart0 code
#define USART0_SUPPORT       (1)
/// non-zero to enable usart1 code
#define USART1_SUPPORT       (0)

/**
 * code is optimized for power of 2 buffer sizes (16, 32, 64, 128, ...)
 * @{
 */
#define USART0_RX_BUFFER_SIZE 16
#define USART0_TX_BUFFER_SIZE 16
#define USART1_RX_BUFFER_SIZE 16
#define USART1_TX_BUFFER_SIZE 16
///@}

/**
 * interrupt driven mode
 *
 * Example:
 * \code
 * #define USART0_INT_MODE
 * #define USART1_INT_MODE
 * // eq
 * #define USART0_TX_INT_MODE
 * #define USART0_RX_INT_MODE
 * #define USART1_TX_INT_MODE
 * #define USART1_RX_INT_MODE
 * \endcode
 * @{
 */
#define USART0_INT_MODE
//#define USART1_INT_MODE
///@}

///@}
//-----------------------------------------------------------------------------

/**
 * use the following macros to determine the 'baud' parameter values
 * for uartInit()
 * \warning 'baud' SHOULD ALWAYS BE A CONSTANT or a lot of code
 * will be generated.
 */
#define USART_BAUD(baud) ((uint16_t)((F_CPU / (16.0 * (baud))) + 0.5) - 1)

/**
 * some typical baud rates
 * confirm the cpu clock rate will support the desired baud rate
 * @{
 */
#define B115200 (USART_BAUD(115200L))
#define B76800  (USART_BAUD(76800UL))
#define B57600  (USART_BAUD(57600UL))
#define B38400  (USART_BAUD(38400UL))
#define B28800  (USART_BAUD(28800U))
#define B19200  (USART_BAUD(19200U))
#define B14400  (USART_BAUD(14400))
#define B9600   (USART_BAUD(9600))
#define B4800   (USART_BAUD(4800))
#define B2400   (USART_BAUD(2400))
///@}

/**
 * Double Speed operation
 * use the following macros to determine the 'baud' parameter values
 * for uartInit() for Double Speed operation.  The macros above will
 * also work just fine.
 * \warning 'baud' SHOULD ALWAYS BE A CONSTANT or a lot of code
 * will be generated.
 * @{
 */
#define DOUBLE_SPEED_BIT (1<<15)
#define USART_2x_BAUD(baud) \
   (((uint16_t)((F_CPU / (8.0 * (baud))) + 0.5) - 1) | DOUBLE_SPEED_BIT)
///@}

/**
 * some typical baud rates
 * confirm the cpu clock rate will support the desired baud rate
 * @{
 */
#define B2x115200 (USART_2x_BAUD(115200L))
#define B2x76800  (USART_2x_BAUD(76800UL))
#define B2x57600  (USART_2x_BAUD(57600UL))
#define B2x38400  (USART_2x_BAUD(38400UL))
#define B2x28800  (USART_2x_BAUD(28800U))
#define B2x19200  (USART_2x_BAUD(19200U))
#define B2x14400  (USART_2x_BAUD(14400))
#define B2x9600   (USART_2x_BAUD(9600))
#define B2x4800   (USART_2x_BAUD(4800))
#define B2x2400   (USART_2x_BAUD(2400))
///@}

#if USART0_SUPPORT
namespace usart0
{
	/**
	 * This function initializes the USART
	 *
	 * \param baud baudrate divisor
	 * \return void
	 *
	 * Example
	 * \code
	 * usart0::init(B9600);
	 * // or
	 * usart0::init(USART_BAUD(9600));
	 * \endcode
	 */
	void init(uint16_t baud);

	/**
	 * This function puts a character into the USART output queue for
	 * transmission.
	 *
	 * \param ch character to be transmitted
	 * \return ch on success, -1 on error (queue full)
	 */
	int16_t putch(char ch);

	/**
	 * This function gets the available space in the transmit queue
	 *
	 * \param void
	 * \return available space in the transmit queue
	 */
	uint16_t space(void);

	/**
	 * This function writes a NULL terminated 'string' to the USART output
	 * queue, returning a pointer to the next character to be written.
	 *
	 * \param *string address of the string
	 * \return a pointer to the next character to be written (\0 if full string is written)
	 */
	const char *puts(const char *string);

	/**
	 * This function writes 'count' characters from 'buffer' to the USART
	 * output queue.
	 *
	 * \param *buffer
	 * \param count
	 * \return 0 on success, -1 if insufficient room, -2 on error
	 * \note if insufficient room, no characters are written.
	 */
	int write(const char *buffer, uint16_t count);

	/**
	 * This function returns the status of the USART transmit data
	 * registers.
	 *
	 * \param void
	 * \return !0 - if both the tx holding & shift registers are empty \
	 *          0 - either the tx holding or shift register is not empty
	 */
	int tx_empty(void);

	/**
	 * This function removes all characters from the USART transmit queue
	 * (without transmitting them).
	 *
	 * \param void
	 * \return void
	 */
	void tx_flush(void);

	/**
	 * This function gets a character from the USART receive queue
	 *
	 * \param void
	 * \return character on success, -1 if no character is available
	 */
	int16_t getch(void);
}
#endif

#if USART1_SUPPORT
namespace usart1
{
	/**
	 * This function initializes the USART1
	 *
	 * \param baud baudrate divisor
	 * \return void
	 * Example:
	 * \code
	 * usart1::init(B9600);
	 * // or
	 * usart1::init(USART_BAUD(9600));
	 * \endcode
	 */
	void init(uint16_t baud);

	/**
	 * This function puts a character into the USART output queue for
	 * transmission.
	 *
	 * \param ch character to be transmitted
	 * \return ch on success, -1 on error (queue full)
	 */
	int16_t putch(char ch);

	/**
	 * This function gets the available space in the transmit queue
	 *
	 * \param void
	 * \return available space in the transmit queue
	 */
	uint16_t space(void);

	/**
	 * This function writes a NULL terminated 'string' to the USART output
	 * queue, returning a pointer to the next character to be written.
	 *
	 * \param *string address of the string
	 * \return a pointer to the next character to be written (\0 if full string is written)
	 */
	const char *puts(const char *string);

	/**
	 * This function writes 'count' characters from 'buffer' to the USART
	 * output queue.
	 *
	 * \param *buffer
	 * \param count
	 * \return 0 on success, -1 if insufficient room, -2 on error
	 * \note if insufficient room, no characters are written.
	 */
	int write(const char *buffer, uint16_t count);

	/**
	 * This function returns the status of the USART transmit data
	 * registers.
	 *
	 * \param void
	 * \return !0 - if both the tx holding & shift registers are empty \
	 *          0 - either the tx holding or shift register is not empty
	 */
	int tx_empty(void);

	/**
	 * This function removes all characters from the USART transmit queue
	 * (without transmitting them).
	 *
	 * \param void
	 * \return void
	 */
	void tx_flush(void);

	/**
	 * This function gets a character from the USART receive queue
	 *
	 * \param void
	 * \return character on success, -1 if no character is available
	 */
	int16_t getch(void);
}
#endif // USART1_SUPPORT
#endif
///@}
